"""
DeiT, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/deit.py

Paper "Training data-efficient image transformers & distillation through attention",
https://arxiv.org/abs/2012.12877
"""

# Reference license: Apache-2.0

import math
from typing import Any
from typing import Optional

import torch
from torch import nn

from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.vit import Encoder
from birder.net.vit import PatchEmbed
from birder.net.vit import adjust_position_embedding


class DeiT(BaseNet):
    block_group_regex = r"encoder\.block\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        attention_dropout = 0.0
        dropout = 0.0
        patch_size: int = self.config["patch_size"]
        num_layers: int = self.config["num_layers"]
        num_heads: int = self.config["num_heads"]
        hidden_dim: int = self.config["hidden_dim"]
        mlp_dim: int = self.config["mlp_dim"]
        drop_path_rate: float = self.config["drop_path_rate"]

        torch._assert(image_size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(image_size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(hidden_dim % num_heads == 0, "Hidden dim indivisible by num heads!")
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.hidden_dim = hidden_dim
        self.num_special_tokens = 2
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]  # Stochastic depth decay rule

        self.conv_proj = nn.Conv2d(
            self.input_channels,
            hidden_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
            bias=True,
        )
        self.patch_embed = PatchEmbed()

        seq_length = (image_size[0] // patch_size) * (image_size[1] // patch_size)

        # Add a class token
        self.class_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
        seq_length += 1

        # Add distillation token
        self.dist_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
        seq_length += 1

        # Add positional embedding
        self.pos_embedding = nn.Parameter(torch.empty(1, seq_length, hidden_dim).normal_(std=0.02))  # from BERT

        self.encoder = Encoder(
            num_layers,
            num_heads,
            hidden_dim,
            mlp_dim,
            dropout,
            attention_dropout,
            dpr,
        )
        self.norm = nn.LayerNorm(hidden_dim, eps=1e-6)

        self.embedding_size = hidden_dim
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()
        self.distillation_output = False

        # Weight initialization
        if isinstance(self.conv_proj, nn.Conv2d):
            # Init the patchify stem
            fan_in = self.conv_proj.in_channels * self.conv_proj.kernel_size[0] * self.conv_proj.kernel_size[1]
            nn.init.trunc_normal_(self.conv_proj.weight, std=math.sqrt(1 / fan_in))
            if self.conv_proj.bias is not None:
                nn.init.zeros_(self.conv_proj.bias)

        if isinstance(self.classifier, nn.Linear):
            nn.init.zeros_(self.classifier.weight)
            nn.init.zeros_(self.classifier.bias)

        if isinstance(self.dist_classifier, nn.Linear):
            nn.init.zeros_(self.dist_classifier.weight)
            nn.init.zeros_(self.dist_classifier.bias)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

            for param in self.dist_classifier.parameters():
                param.requires_grad = True

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Expand the class token to the full batch
        batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
        batch_dist_token = self.dist_token.expand(x.shape[0], -1, -1)

        x = torch.concat([batch_class_token, batch_dist_token, x], dim=1)
        x = x + self.pos_embedding

        x = self.encoder(x)
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        x = x[:, 0:2]

        return x

    def set_distillation_output(self, enable: bool = True) -> None:
        self.distillation_output = enable

    def classify(self, x: torch.Tensor) -> torch.Tensor:
        x_cls = x[:, 0]
        x_dist = x[:, 1]

        x_cls = self.classifier(x_cls)
        x_dist = self.dist_classifier(x_dist)

        if self.training is True and self.distillation_output is True:
            x = torch.stack([x_cls, x_dist], dim=1)
        else:
            # Classifier "token" as an average of both tokens (during regular training or inference)
            x = (x_cls + x_dist) / 2

        return x

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        old_size = self.size
        super().adjust_size(new_size)

        # Sort out sizes
        num_prefix_tokens = 2

        # Add back class tokens
        self.pos_embedding = nn.Parameter(
            adjust_position_embedding(
                self.pos_embedding,
                (old_size[0] // self.patch_size, old_size[1] // self.patch_size),
                (new_size[0] // self.patch_size, new_size[1] // self.patch_size),
                num_prefix_tokens,
            )
        )


registry.register_model_config(
    "deit_t16",
    DeiT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 3,
        "hidden_dim": 192,
        "mlp_dim": 768,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "deit_s16",
    DeiT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "deit_b16",
    DeiT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "deit_t16_il-common",
    {
        "description": "DeiT tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 21.7,
                "sha256": "ac124122dec9f1bceff383a6a555ca375ca1b613caf486dac3f29d87afac03b3",
            }
        },
        "net": {"network": "deit_t16", "tag": "il-common"},
    },
)
registry.register_weights(
    "deit_t16_dist-il-common",
    {
        "description": (
            "DeiT tiny model trained using distillation from a ConvNeXt v2 tiny teacher on the il-common dataset"
        ),
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 21.7,
                "sha256": "fafd0c3c65f9c35318f449f60485f640917736ee7b44056be55c2226909ffdb8",
            }
        },
        "net": {"network": "deit_t16", "tag": "dist-il-common"},
    },
)



"""
EfficientViT (Microsoft Research), adapted from
https://github.com/microsoft/Cream/blob/main/EfficientViT/classification/model/efficientvit.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/efficientvit_msra.py

Paper "EfficientViT: Memory Efficient Vision Transformer with Cascaded Group Attention",
https://arxiv.org/abs/2305.07027

Changes from original:
* Remove batchnorm fuse
* Window size based on image size (image size // 32)
"""

# Reference license: MIT and Apache-2.0

import itertools
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import SqueezeExcitation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import interpolate_attention_bias


class Conv2dNorm(nn.Sequential):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        groups: int = 1,
        bn_weight_init: float = 1.0,
    ) -> None:
        super().__init__()
        self.add_module(
            "conv",
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=groups,
                bias=False,
            ),
        )
        self.add_module("bn", nn.BatchNorm2d(out_channels))
        nn.init.constant_(self.bn.weight, bn_weight_init)
        nn.init.constant_(self.bn.bias, 0)


class PatchMerging(nn.Module):
    def __init__(self, dim: int, out_dim: int) -> None:
        super().__init__()
        hidden_dim = dim * 4
        self.conv1 = Conv2dNorm(dim, hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.ReLU()
        self.conv2 = Conv2dNorm(
            hidden_dim, hidden_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), groups=hidden_dim
        )
        self.se = SqueezeExcitation(hidden_dim, hidden_dim // 4)
        self.conv3 = Conv2dNorm(hidden_dim, out_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.act(x)
        x = self.conv2(x)
        x = self.act(x)
        x = self.se(x)
        x = self.conv3(x)

        return x


class ResidualDrop(nn.Module):
    def __init__(self, m: nn.Module, drop: float) -> None:
        super().__init__()
        self.m = m
        self.drop = drop

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.training is True and self.drop > 0:
            return (
                x
                + self.m(x) * torch.rand(x.size(0), 1, 1, 1, device=x.device).ge_(self.drop).div(1 - self.drop).detach()
            )

        return x + self.m(x)


class FFN(nn.Module):
    def __init__(self, dim: int, hidden_dim: int) -> None:
        super().__init__()
        self.pw1 = Conv2dNorm(dim, hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.ReLU()
        self.pw2 = Conv2dNorm(hidden_dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bn_weight_init=0.0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pw1(x)
        x = self.act(x)
        x = self.pw2(x)

        return x


class CascadedGroupAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        key_dim: int,
        num_heads: int,
        attn_ratio: float,
        resolution: tuple[int, int],
        kernels: list[int],
    ) -> None:
        super().__init__()
        self.scale = key_dim**-0.5
        self.key_dim = key_dim
        self.d = int(attn_ratio * key_dim)

        qkvs = []
        dws = []
        for i in range(num_heads):
            qkvs.append(
                Conv2dNorm(
                    dim // (num_heads), self.key_dim * 2 + self.d, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
                )
            )
            dws.append(
                Conv2dNorm(
                    self.key_dim,
                    self.key_dim,
                    kernel_size=(kernels[i], kernels[i]),
                    stride=(1, 1),
                    padding=(kernels[i] // 2, kernels[i] // 2),
                    groups=self.key_dim,
                )
            )

        self.qkvs = nn.ModuleList(qkvs)
        self.dws = nn.ModuleList(dws)
        self.proj = nn.Sequential(
            nn.ReLU(),
            Conv2dNorm(self.d * num_heads, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bn_weight_init=0.0),
        )

        points = list(itertools.product(range(resolution[0]), range(resolution[1])))
        N = len(points)
        attention_offsets: dict[tuple[int, int], int] = {}
        idxs = []
        for p1 in points:
            for p2 in points:
                offset = (abs(p1[0] - p2[0]), abs(p1[1] - p2[1]))
                if offset not in attention_offsets:
                    attention_offsets[offset] = len(attention_offsets)

                idxs.append(attention_offsets[offset])

        self.attention_biases = nn.Parameter(torch.zeros(num_heads, len(attention_offsets)))
        self.attention_bias_idxs = nn.Buffer(torch.LongTensor(idxs).view(N, N), persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, _, H, W) = x.size()
        feats_in = x.chunk(len(self.qkvs), dim=1)
        feats_out = []
        feat = feats_in[0]
        attn_bias = self.attention_biases[:, self.attention_bias_idxs]
        for head_idx, (qkv, dws) in enumerate(zip(self.qkvs, self.dws)):
            if head_idx > 0:
                feat = feat + feats_in[head_idx]

            feat = qkv(feat)
            (q, k, v) = feat.view(B, -1, H, W).split([self.key_dim, self.key_dim, self.d], dim=1)
            q = dws(q)
            q = q.flatten(2)
            k = k.flatten(2)
            v = v.flatten(2)

            q = q * self.scale
            attn = q.transpose(-2, -1) @ k
            attn = attn + attn_bias[head_idx]  # pylint: disable=unsubscriptable-object
            attn = attn.softmax(dim=-1)

            feat = v @ attn.transpose(-2, -1)
            feat = feat.view(B, self.d, H, W)
            feats_out.append(feat)

        x = torch.concat(feats_out, dim=1)
        x = self.proj(x)

        return x


class LocalWindowAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        key_dim: int,
        num_heads: int,
        attn_ratio: float,
        resolution: tuple[int, int],
        window_resolution: tuple[int, int],
        kernels: list[int],
    ) -> None:
        super().__init__()
        self.resolution = resolution
        assert window_resolution[0] > 0 and window_resolution[1] > 0, "window_size must be greater than 0"

        self.window_resolution = window_resolution
        window_resolution = (min(window_resolution[0], resolution[0]), min(window_resolution[1], resolution[1]))
        self.attn = CascadedGroupAttention(
            dim,
            key_dim,
            num_heads,
            attn_ratio=attn_ratio,
            resolution=window_resolution,
            kernels=kernels,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        H = self.resolution[0]
        W = self.resolution[1]
        (B, C, _, _) = x.size()

        if H <= self.window_resolution[0] and W <= self.window_resolution[1]:
            x = self.attn(x)
        else:
            x = x.permute(0, 2, 3, 1)
            pad_b = (self.window_resolution[0] - H % self.window_resolution[0]) % self.window_resolution[0]
            pad_r = (self.window_resolution[1] - W % self.window_resolution[1]) % self.window_resolution[1]
            x = F.pad(x, (0, 0, 0, pad_r, 0, pad_b))

            p_h = H + pad_b
            p_w = W + pad_r
            n_h = p_h // self.window_resolution[0]
            n_w = p_w // self.window_resolution[1]

            # Window partition, BHWC -> B(nHh)(nWw)C -> BnHnWhwC -> (BnHnW)hwC -> (BnHnW)Chw
            x = x.view(B, n_h, self.window_resolution[0], n_w, self.window_resolution[1], C).transpose(2, 3)
            x = x.reshape(B * n_h * n_w, self.window_resolution[0], self.window_resolution[1], C).permute(0, 3, 1, 2)
            x = self.attn(x)

            # Window reverse, (BnHnW)Chw -> (BnHnW)hwC -> BnHnWhwC -> B(nHh)(nWw)C -> BHWC
            x = x.permute(0, 2, 3, 1).view(B, n_h, n_w, self.window_resolution[0], self.window_resolution[1], C)
            x = x.transpose(2, 3).reshape(B, p_h, p_w, C)
            x = x[:, :H, :W].contiguous()
            x = x.permute(0, 3, 1, 2)

        return x


class EfficientVitBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        key_dim: int,
        num_heads: int,
        attn_ratio: float,
        resolution: tuple[int, int],
        window_resolution: tuple[int, int],
        kernels: list[int],
    ) -> None:
        super().__init__()

        self.dw0 = ResidualDrop(
            Conv2dNorm(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim, bn_weight_init=0.0),
            drop=0.0,
        )
        self.ffn0 = ResidualDrop(FFN(dim, int(dim * 2)), drop=0.0)

        self.mixer = ResidualDrop(
            LocalWindowAttention(
                dim,
                key_dim,
                num_heads,
                attn_ratio=attn_ratio,
                resolution=resolution,
                window_resolution=window_resolution,
                kernels=kernels,
            ),
            drop=0.0,
        )

        self.dw1 = ResidualDrop(
            Conv2dNorm(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim, bn_weight_init=0.0),
            drop=0.0,
        )
        self.ffn1 = ResidualDrop(FFN(dim, int(dim * 2)), drop=0.0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.dw0(x)
        x = self.ffn0(x)
        x = self.mixer(x)
        x = self.dw1(x)
        x = self.ffn1(x)

        return x


class EfficientVitStage(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        key_dim: int,
        downsample: bool,
        num_heads: int,
        attn_ratio: float,
        resolution: tuple[int, int],
        window_resolution: tuple[int, int],
        kernels: list[int],
        depth: int,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.resolution = ((resolution[0] - 1) // 2 + 1, (resolution[1] - 1) // 2 + 1)
            self.downsample = nn.Sequential(
                ResidualDrop(
                    Conv2dNorm(in_dim, in_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=in_dim),
                    drop=0.0,
                ),
                ResidualDrop(FFN(in_dim, int(in_dim * 2)), drop=0.0),
                PatchMerging(in_dim, out_dim),
                ResidualDrop(
                    Conv2dNorm(out_dim, out_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=out_dim),
                    drop=0.0,
                ),
                ResidualDrop(FFN(out_dim, int(out_dim * 2)), drop=0.0),
            )

        else:
            assert in_dim == out_dim
            self.resolution = resolution
            self.downsample = nn.Identity()

        blocks = []
        for _ in range(depth):
            blocks.append(
                EfficientVitBlock(out_dim, key_dim, num_heads, attn_ratio, self.resolution, window_resolution, kernels)
            )

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


# pylint: disable=invalid-name
class EfficientViT_MSFT(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        img_size = self.size
        key_dims = [16, 16, 16]
        embed_dims: list[int] = self.config["embed_dims"]
        depths: list[int] = self.config["depths"]
        num_heads: list[int] = self.config["num_heads"]
        kernels: list[int] = self.config["kernels"]
        window_size = [(int(img_size[0] / (2**5)), int(img_size[1] / (2**5)))] * len(depths)
        resolution = (img_size[0] // 16, img_size[1] // 16)
        attn_ratio = [embed_dims[i] / (key_dims[i] * num_heads[i]) for i in range(len(embed_dims))]

        self.stem = nn.Sequential(
            Conv2dNorm(self.input_channels, embed_dims[0] // 8, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            nn.ReLU(),
            Conv2dNorm(embed_dims[0] // 8, embed_dims[0] // 4, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            nn.ReLU(),
            Conv2dNorm(embed_dims[0] // 4, embed_dims[0] // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            nn.ReLU(),
            Conv2dNorm(embed_dims[0] // 2, embed_dims[0], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )

        num_stages = len(depths)
        prev_dim = embed_dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stage = EfficientVitStage(
                in_dim=prev_dim,
                out_dim=embed_dims[i],
                key_dim=key_dims[i],
                downsample=i > 0,
                num_heads=num_heads[i],
                attn_ratio=attn_ratio[i],
                resolution=resolution,
                window_resolution=window_size[i],
                kernels=kernels,
                depth=depths[i],
            )
            stages[f"stage{i+1}"] = stage
            return_channels.append(embed_dims[i])
            resolution = stage.resolution
            prev_dim = embed_dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
            nn.BatchNorm1d(embed_dims[-1]),
        )
        self.return_stages = self.return_stages[: len(depths)]
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        old_resolution = (old_size[0] // 16, old_size[1] // 16)
        old_window_size = (int(old_size[0] / (2**5)), int(old_size[1] / (2**5)))
        resolution = (new_size[0] // 16, new_size[1] // 16)
        window_size = (int(new_size[0] / (2**5)), int(new_size[1] / (2**5)))
        for module in self.body.children():  # pylint: disable=too-many-nested-blocks
            if isinstance(module, EfficientVitStage):
                if not isinstance(module.downsample, nn.Identity):
                    old_resolution = ((old_resolution[0] - 1) // 2 + 1, (old_resolution[1] - 1) // 2 + 1)
                    resolution = ((resolution[0] - 1) // 2 + 1, (resolution[1] - 1) // 2 + 1)

                module.resolution = resolution
                for m in module.modules():
                    if isinstance(m, EfficientVitBlock):
                        m.mixer.m.resolution = resolution
                        m.mixer.m.window_resolution = window_size
                        old_window_resolution = (
                            min(old_window_size[0], old_resolution[0]),
                            min(old_window_size[1], old_resolution[1]),
                        )
                        window_resolution = (min(window_size[0], resolution[0]), min(window_size[1], resolution[1]))

                        # Interpolate attention_biases
                        points = list(itertools.product(range(window_resolution[0]), range(window_resolution[1])))
                        N = len(points)
                        attention_offsets: dict[tuple[int, int], int] = {}
                        idxs = []
                        for p1 in points:
                            for p2 in points:
                                offset = (abs(p1[0] - p2[0]), abs(p1[1] - p2[1]))
                                if offset not in attention_offsets:
                                    attention_offsets[offset] = len(attention_offsets)

                                idxs.append(attention_offsets[offset])

                        m.mixer.m.attn.attention_biases = nn.Parameter(
                            interpolate_attention_bias(
                                m.mixer.m.attn.attention_biases, old_window_resolution, window_resolution
                            )
                        )
                        m.mixer.m.attn.attention_bias_idxs = nn.Buffer(
                            torch.LongTensor(idxs).view(N, N), persistent=False
                        )


registry.register_model_config(
    "efficientvit_msft_m0",
    EfficientViT_MSFT,
    config={"embed_dims": [64, 128, 192], "depths": [1, 2, 3], "num_heads": [4, 4, 4], "kernels": [5, 5, 5, 5]},
)
registry.register_model_config(
    "efficientvit_msft_m1",
    EfficientViT_MSFT,
    config={"embed_dims": [128, 144, 192], "depths": [1, 2, 3], "num_heads": [2, 3, 3], "kernels": [7, 5, 3, 3]},
)
registry.register_model_config(
    "efficientvit_msft_m2",
    EfficientViT_MSFT,
    config={"embed_dims": [128, 192, 224], "depths": [1, 2, 3], "num_heads": [4, 3, 2], "kernels": [7, 5, 3, 3]},
)
registry.register_model_config(
    "efficientvit_msft_m3",
    EfficientViT_MSFT,
    config={"embed_dims": [128, 240, 320], "depths": [1, 2, 3], "num_heads": [4, 3, 4], "kernels": [5, 5, 5, 5]},
)
registry.register_model_config(
    "efficientvit_msft_m4",
    EfficientViT_MSFT,
    config={"embed_dims": [128, 256, 384], "depths": [1, 2, 3], "num_heads": [4, 4, 4], "kernels": [7, 5, 3, 3]},
)
registry.register_model_config(
    "efficientvit_msft_m5",
    EfficientViT_MSFT,
    config={"embed_dims": [192, 288, 384], "depths": [1, 3, 4], "num_heads": [3, 3, 4], "kernels": [7, 5, 3, 3]},
)

registry.register_weights(
    "efficientvit_msft_m0_il-common",
    {
        "description": "EfficientViT (MSFT) M0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 8.9,
                "sha256": "f1ff44ac458b72ccdf442ce5516ee01e305d0ba3123e5512a5b1579a9ae8d422",
            }
        },
        "net": {"network": "efficientvit_msft_m0", "tag": "il-common"},
    },
)



"""
RegNet Z, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/regnet.py

Paper "Fast and Accurate Model Scaling", https://arxiv.org/abs/2103.06877
"""

# Reference license: Apache-2.0

import math
from collections import OrderedDict
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.regnet import BlockParams
from birder.net.regnet import BottleneckTransform


class ResBottleneckBlock(nn.Module):
    def __init__(
        self,
        width_in: int,
        width_out: int,
        stride: tuple[int, int],
        group_width: int,
        bottleneck_multiplier: float,
        se_ratio: float,
    ) -> None:
        super().__init__()

        if width_in != width_out or stride[0] != 1 or stride[1] != 1:
            self.shortcut = False

        else:
            self.shortcut = True

        self.f = BottleneckTransform(
            width_in,
            width_out,
            stride=stride,
            group_width=group_width,
            bottleneck_multiplier=bottleneck_multiplier,
            se_ratio=se_ratio,
            activation_layer=nn.SiLU,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.shortcut is True:
            return x + self.f(x)

        return self.f(x)


class AnyStage(nn.Module):
    def __init__(
        self,
        width_in: int,
        width_out: int,
        stride: tuple[int, int],
        depth: int,
        group_width: int,
        bottleneck_multiplier: float,
        se_ratio: float,
    ) -> None:
        super().__init__()

        layers = []
        for i in range(depth):
            if i == 0:
                in_ch = width_in
                cur_stride = stride
            else:
                in_ch = width_out
                cur_stride = (1, 1)

            layers.append(
                ResBottleneckBlock(
                    in_ch,
                    width_out,
                    stride=cur_stride,
                    group_width=group_width,
                    bottleneck_multiplier=bottleneck_multiplier,
                    se_ratio=se_ratio,
                )
            )

        self.block = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.block(x)


# pylint: disable=invalid-name
class RegNet_Z(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.block\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        stem_width = 32
        bottleneck_multiplier = 4.0
        se_ratio = 0.25
        depth: int = self.config["depth"]
        w_0: int = self.config["w_0"]
        w_a: float = self.config["w_a"]
        w_m: float = self.config["w_m"]
        group_width: int = self.config["group_width"]
        num_features: int = self.config["num_features"]

        block_params = BlockParams.from_init_params(depth, w_0, w_a, w_m, group_width, bottleneck_multiplier, se_ratio)

        self.stem = Conv2dNormActivation(
            self.input_channels,
            stem_width,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            activation_layer=nn.SiLU,
            bias=False,
        )

        current_width = stem_width
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        i = 0
        for (
            width_out,
            stride,
            depth,
            group_width,
            bottleneck_multiplier,
        ) in block_params._get_expanded_params():
            stages[f"stage{i+1}"] = AnyStage(
                current_width, width_out, (stride, stride), depth, group_width, bottleneck_multiplier, se_ratio
            )
            return_channels.append(width_out)
            current_width = width_out
            i += 1

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                current_width, num_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=nn.SiLU
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = num_features
        self.classifier = self.create_classifier()

        self.stem_stride = 2
        self.stem_width = stem_width
        self.encoding_size = current_width
        decoder_block = partial(
            BottleneckTransform,
            stride=(1, 1),
            group_width=64,
            bottleneck_multiplier=1.0,
            se_ratio=se_ratio,
            activation_layer=nn.SiLU,
        )
        self.decoder_block = lambda x: decoder_block(x, x)

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                # Note that there is no bias due to BN
                fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                nn.init.normal_(m.weight, mean=0.0, std=math.sqrt(2.0 / fan_out))

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, mean=0.0, std=0.01)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "regnet_z_500m",
    RegNet_Z,
    config={"depth": 21, "w_0": 16, "w_a": 10.7, "w_m": 2.51, "group_width": 4, "num_features": 1024},
)
registry.register_model_config(
    "regnet_z_4g",
    RegNet_Z,
    config={"depth": 28, "w_0": 48, "w_a": 14.5, "w_m": 2.226, "group_width": 8, "num_features": 1536},
)

registry.register_weights(
    "regnet_z_500m_il-common",
    {
        "description": "RegNet Z 500m model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 25.1,
                "sha256": "b2cc5c9f5c5e4693d8fe12e2d5eddaa28ce25d9ea38e14ea67ec09706aa24ea9",
            }
        },
        "net": {"network": "regnet_z_500m", "tag": "il-common"},
    },
)



import random
from collections.abc import Callable
from typing import Any
from typing import Literal
from typing import Optional
from typing import TypedDict

import torch
from torch import nn
from torchvision.transforms import v2
from torchvision.transforms.v2 import functional as F

RGBType = TypedDict("RGBType", {"mean": tuple[float, float, float], "std": tuple[float, float, float]})
RGBMode = Literal["birder", "imagenet", "clip", "neutral", "none"]


def get_rgb_stats(
    mode: RGBMode, mean: Optional[tuple[float, float, float]] = None, std: Optional[tuple[float, float, float]] = None
) -> RGBType:
    if mode == "birder":
        default_mean = (0.5191, 0.5306, 0.4877)
        default_std = (0.2316, 0.2304, 0.2588)
    elif mode == "imagenet":
        default_mean = (0.485, 0.456, 0.406)
        default_std = (0.229, 0.224, 0.225)
    elif mode == "clip":  # OpenAI CLIP - https://github.com/openai/CLIP/blob/main/clip/clip.py#L85
        default_mean = (0.48145466, 0.4578275, 0.40821073)
        default_std = (0.26862954, 0.26130258, 0.27577711)
    elif mode == "neutral":
        default_mean = (0.0, 0.0, 0.0)
        default_std = (1.0, 1.0, 1.0)
    elif mode == "none":
        default_mean = (0.5, 0.5, 0.5)
        default_std = (0.5, 0.5, 0.5)
    else:
        raise ValueError(f"unknown mode={mode}")

    return {
        "mean": mean if mean is not None else default_mean,
        "std": std if std is not None else default_std,
    }


def get_mixup_cutmix(alpha: Optional[float], num_outputs: int, cutmix: bool) -> Callable[..., torch.Tensor]:
    choices: list[Callable[..., torch.Tensor]] = []
    choices.append(v2.Identity())
    if alpha is not None:
        choices.append(v2.MixUp(alpha=alpha, num_classes=num_outputs))

    if cutmix is True:
        choices.append(v2.CutMix(alpha=1.0, num_classes=num_outputs))

    return v2.RandomChoice(choices)  # type: ignore


# Using transforms v2 mixup, keeping this implementation only as a reference
class RandomMixup(nn.Module):
    """
    Randomly apply Mixup to the provided batch and targets.

    The class implements the data augmentations as described in the paper
    "mixup: Beyond Empirical Risk Minimization"
    https://arxiv.org/abs/1710.09412

    Parameters
    ----------
    num_classes
        Number of classes used for one-hot encoding.
    p
        Probability of the batch being transformed
    alpha
        Hyperparameter of the Beta distribution used for mixup.
    """

    def __init__(self, num_classes: int, p: float, alpha: float) -> None:
        super().__init__()

        if num_classes < 1:
            raise ValueError(
                f"Please provide a valid positive value for the num_classes. Got num_classes={num_classes}"
            )

        if alpha <= 0:
            raise ValueError("Alpha param must be positive")

        self.num_classes = num_classes
        self.p = p
        self.alpha = alpha

    def forward(self, batch: torch.Tensor, target: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Parameters
        ----------
        batch
            Float tensor of size (B, C, H, W)
        target
            Integer tensor of size (B, )

        Returns
        -------
        Randomly transformed batch.

        Raises
        ------
        ValueError
            On wrong tensor dimensions.
        TypeError
            On bad tensor dtype.
        """

        if batch.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {batch.ndim}")

        if target.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target.ndim}")

        if batch.is_floating_point() is False:
            raise TypeError(f"Batch dtype should be a float tensor. Got {batch.dtype}.")

        if target.dtype != torch.int64:
            raise TypeError(f"Target dtype should be torch.int64. Got {target.dtype}")

        batch = batch.clone()
        target = target.clone()

        if target.ndim == 1:
            # pylint: disable=not-callable
            target = nn.functional.one_hot(target, num_classes=self.num_classes).to(dtype=batch.dtype)

        if torch.rand(1).item() >= self.p:
            return (batch, target)

        # It's faster to roll the batch by one instead of shuffling it to create image pairs
        batch_rolled = batch.roll(1, 0)
        target_rolled = target.roll(1, 0)

        # Implemented as on mixup paper, page 3.
        lambda_param = float(
            torch._sample_dirichlet(torch.tensor([self.alpha, self.alpha]))[0]  # pylint: disable=protected-access
        )
        batch_rolled.mul_(1.0 - lambda_param)
        batch.mul_(lambda_param).add_(batch_rolled)

        target_rolled.mul_(1.0 - lambda_param)
        target.mul_(lambda_param).add_(target_rolled)

        return (batch, target)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(num_classes={self.num_classes}, p={self.p}, alpha={self.alpha})"


class RandomResizedCropWithRandomInterpolation(nn.Module):
    def __init__(
        self,
        size: tuple[int, int],
        scale: tuple[float, float],
        ratio: tuple[float, float],
        interpolation: list[v2.InterpolationMode],
    ) -> None:
        super().__init__()
        self.transform = []
        for interp in interpolation:
            self.transform.append(
                v2.RandomResizedCrop(
                    size,
                    scale=scale,
                    ratio=ratio,
                    interpolation=interp,
                    antialias=True,
                )
            )

    def forward(self, x: Any) -> torch.Tensor:
        t = random.choice(self.transform)
        return t(x)


class SimpleRandomCropWithRandomInterpolation(nn.Module):
    def __init__(self, size: tuple[int, int], interpolation: list[v2.InterpolationMode]) -> None:
        super().__init__()
        self.transform = []
        for interp in interpolation:
            self.transform.append(
                v2.Compose(
                    [
                        v2.Resize(min(size), interpolation=interp),
                        v2.RandomCrop(size, padding=4, padding_mode="reflect"),
                    ]
                )
            )

    def forward(self, x: Any) -> torch.Tensor:
        t = random.choice(self.transform)
        return t(x)


class RandomSolarizeWithVariableThreshold(nn.Module):
    def __init__(self, threshold: int, p: float, threshold_variation: float) -> None:
        super().__init__()
        self.threshold = threshold
        self.p = p
        self.threshold_variation = threshold_variation

    def forward(self, x: Any) -> torch.Tensor:
        if torch.rand(1) >= self.p:
            return x

        noise_range = int(self.threshold * self.threshold_variation)
        noise = torch.randint(-noise_range, noise_range + 1, (1,)).item()
        threshold = max(0, min(255, self.threshold + noise))

        return F.solarize(x, threshold=threshold)


class RandomAdjustSharpnessWithVariation(nn.Module):
    def __init__(self, factor: float, p: float, factor_variation: float) -> None:
        super().__init__()
        self.factor = factor
        self.p = p
        self.factor_variation = factor_variation

    def forward(self, x: Any) -> torch.Tensor:
        if torch.rand(1) >= self.p:
            return x

        noise = torch.rand((1,)).item()
        noise = noise * 2 * self.factor_variation - self.factor_variation
        factor = max(1.0, min(2.0, self.factor + noise))

        return F.adjust_sharpness(x, factor)


class RandomSaturationWithVariation(nn.Module):
    def __init__(self, factor: float, p: float, factor_variation: float) -> None:
        super().__init__()
        self.factor = factor
        self.p = p
        self.factor_variation = factor_variation

    def forward(self, x: Any) -> torch.Tensor:
        if torch.rand(1) >= self.p:
            return x

        noise = torch.rand((1,)).item()
        noise = noise * 2 * self.factor_variation - self.factor_variation
        factor = max(0.1, min(2.0, self.factor + noise))

        return F.adjust_saturation(x, factor)


class BirderAugment(nn.Module):
    def __init__(self, level: int, re_prob: Optional[float] = None, use_grayscale: bool = False):
        super().__init__()
        assert level <= 10

        self.min_ops = 2
        self.max_ops = max(self.min_ops, (level + 1) // 2)

        re_scale = 0.05 + (level * 0.025)
        if re_prob is None:
            re_prob = max(0.25, -0.15 + (level * 0.05))

        self.re = v2.Identity() if re_prob < 0.001 else v2.RandomErasing(re_prob, scale=(0.02, re_scale))

        self.transformations = []
        self.transformations.append(v2.Identity())

        if level >= 1:
            self.transformations.extend(
                [
                    v2.RandomAffine(degrees=2.25 * level),
                    v2.ColorJitter(brightness=0.2 + (0.0125 * level)),
                    v2.ColorJitter(contrast=0.05 + (0.025 * level), hue=max(0, -0.025 + (level * 0.0125))),
                ]
            )
        if level >= 3:
            self.transformations.extend(
                [
                    RandomSaturationWithVariation(1.0, p=1.0, factor_variation=level * 0.07),
                    v2.RandomAffine(degrees=0, translate=(0.02 * level, 0.02 * level)),
                    v2.RandomAffine(degrees=0, shear=(-2.5 * level, 2.5 * level, 0, 0)),
                    v2.RandomPosterize(9 - ((level + 1) // 2), p=1.0),
                    v2.GaussianBlur(kernel_size=((level + 1) // 4) * 2 + 1, sigma=(0.5, 1.2)),
                    v2.RandomAutocontrast(1.0),
                    RandomAdjustSharpnessWithVariation(1 + (level * 0.04), p=1.0, factor_variation=0.2),
                ]
            )
        if level >= 5:
            self.transformations.append(
                RandomSolarizeWithVariableThreshold(255 - (12 * level), p=1.0, threshold_variation=0.3)
            )

        if use_grayscale is True:
            self.transformations.append(v2.RandomGrayscale(1.0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        num_ops = torch.randint(self.min_ops, self.max_ops + 1, ()).item()
        for _ in range(num_ops):
            t = self.transformations[torch.randint(len(self.transformations), ()).item()]
            x = t(x)

        x = self.re(x)

        return x


AugType = Literal["birder", "aa", "ra", "ta_wide", "augmix", "3aug"]


# pylint: disable=too-many-branches
def training_preset(
    size: tuple[int, int],
    aug_type: AugType,
    level: int,
    rgv_values: RGBType,
    resize_min_scale: Optional[float] = None,
    re_prob: Optional[float] = None,
    use_grayscale: bool = False,
    ra_num_ops: int = 2,
    ra_magnitude: int = 9,
    augmix_severity: int = 3,
    simple_crop: bool = False,
) -> Callable[..., torch.Tensor]:
    mean = rgv_values["mean"]
    std = rgv_values["std"]

    if aug_type == "birder":
        if 0 > level or level > 10:
            raise ValueError("Unsupported aug level")

        if level == 0:
            return v2.Compose(  # type: ignore
                [
                    v2.Resize(size, interpolation=v2.InterpolationMode.BICUBIC, antialias=True),
                    v2.PILToTensor(),
                    v2.ToDtype(torch.float32, scale=True),
                    v2.Normalize(mean=mean, std=std),
                ]
            )

        if resize_min_scale is None:
            resize_min_scale = 0.8 - (level * 0.05)

        if simple_crop is True:
            crop_transform = SimpleRandomCropWithRandomInterpolation(
                size, interpolation=[v2.InterpolationMode.BILINEAR, v2.InterpolationMode.BICUBIC]
            )
        else:
            crop_transform = RandomResizedCropWithRandomInterpolation(
                size,
                scale=(resize_min_scale, 1.0),
                ratio=(3 / 4, 4 / 3),
                interpolation=[v2.InterpolationMode.BILINEAR, v2.InterpolationMode.BICUBIC],
            )

        return v2.Compose(  # type:ignore
            [
                v2.PILToTensor(),
                crop_transform,
                BirderAugment(level, re_prob, use_grayscale),
                v2.RandomHorizontalFlip(0.5),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
            ]
        )

    if resize_min_scale is None:
        resize_min_scale = 0.08
    if re_prob is None:
        re_prob = 0.0

    transforms = [v2.PILToTensor()]
    if simple_crop is True:
        transforms.append(
            SimpleRandomCropWithRandomInterpolation(
                size, interpolation=[v2.InterpolationMode.BILINEAR, v2.InterpolationMode.BICUBIC]
            )
        )
    else:
        transforms.append(
            RandomResizedCropWithRandomInterpolation(
                size,
                scale=(resize_min_scale, 1.0),
                ratio=(3 / 4, 4 / 3),
                interpolation=[v2.InterpolationMode.BILINEAR, v2.InterpolationMode.BICUBIC],
            )
        )

    if aug_type == "aa":  # AutoAugment policy
        transforms.append(v2.AutoAugment(v2.AutoAugmentPolicy.IMAGENET, v2.InterpolationMode.BILINEAR))
    elif aug_type == "ra":  # RandAugment policy
        transforms.append(v2.RandAugment(ra_num_ops, ra_magnitude, interpolation=v2.InterpolationMode.BILINEAR))
    elif aug_type == "ta_wide":  # TrivialAugmentWide policy
        transforms.append(v2.TrivialAugmentWide(interpolation=v2.InterpolationMode.BILINEAR))
    elif aug_type == "augmix":
        transforms.append(v2.AugMix(augmix_severity, interpolation=v2.InterpolationMode.BILINEAR))
    elif aug_type == "3aug":
        transforms.append(
            v2.RandomChoice(
                [v2.RandomGrayscale(p=1.0), v2.RandomSolarize(128, p=1.0), v2.GaussianBlur(kernel_size=(3, 3))]
            )
        )
        transforms.append(v2.ColorJitter(brightness=0.3, contrast=0.3, hue=0.3))
    else:
        raise ValueError("Unsupported augmentation type")

    return v2.Compose(  # type:ignore
        [
            *transforms,
            v2.RandomHorizontalFlip(0.5),
            v2.Identity() if re_prob == 0 else v2.RandomErasing(re_prob),
            v2.ToDtype(torch.float32, scale=True),
            v2.Normalize(mean=mean, std=std),
        ]
    )


def inference_preset(
    size: tuple[int, int], rgv_values: RGBType, center_crop: float = 1.0, simple_crop: bool = False
) -> Callable[..., torch.Tensor]:
    mean = rgv_values["mean"]
    std = rgv_values["std"]

    if simple_crop is True:
        base_size: int | tuple[int, int] = int(min(size) / center_crop)
    else:
        base_size = (int(size[0] / center_crop), int(size[1] / center_crop))

    return v2.Compose(  # type: ignore
        [
            v2.Resize(base_size, interpolation=v2.InterpolationMode.BICUBIC, antialias=True),
            v2.CenterCrop(size),
            v2.PILToTensor(),
            v2.ToDtype(torch.float32, scale=True),
            v2.Normalize(mean=mean, std=std),
        ]
    )


def reverse_preset(rgv_values: RGBType) -> Callable[..., torch.Tensor]:
    mean = rgv_values["mean"]
    std = rgv_values["std"]

    reverse_mean = [-m / s for m, s in zip(mean, std)]
    reverse_std = [1 / s for s in std]

    return v2.Compose(  # type: ignore
        [
            v2.Normalize(mean=reverse_mean, std=reverse_std),
            v2.ToDtype(torch.uint8, scale=True),
        ]
    )



import logging
import math
from collections.abc import Callable
from typing import Any
from typing import Optional

import webdataset as wds
from torch.utils.data import DataLoader

logger = logging.getLogger(__name__)


def make_wds_loader(
    dataset: wds.WebDataset,
    batch_size: int,
    num_workers: int,
    prefetch_factor: Optional[int],
    collate_fn: Optional[Callable[..., Any]],
    world_size: int,
    pin_memory: bool,
    drop_last: bool = False,
    shuffle: bool = False,
    *,
    exact: bool = False,
) -> DataLoader:
    dataloader = wds.WebLoader(
        dataset.repeat() if exact is False else dataset,
        batch_size=batch_size,
        num_workers=num_workers,
        prefetch_factor=prefetch_factor,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
        drop_last=drop_last,
    )
    if shuffle is True:
        logger.info("WDS extra shuffle enabled: applying global batch-level shuffling")
        dataloader = dataloader.unbatched().shuffle(1000).batched(batch_size)

    dataloader.batch_size = batch_size
    if drop_last is True:
        epoch_size = math.floor(len(dataset) / (batch_size * world_size))
    else:
        epoch_size = math.ceil(len(dataset) / (batch_size * world_size))

    dataloader = dataloader.with_length(epoch_size, silent=True)
    if exact is False:
        dataloader = dataloader.with_epoch(epoch_size)

    return dataloader



"""
EfficientViT (MIT), adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/efficientvit_mit.py

Paper "EfficientViT: Multi-Scale Linear Attention for High-Resolution Dense Prediction",
https://arxiv.org/abs/2205.14756
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.layers.activations import get_activation_module
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class ResidualBlock(nn.Module):
    def __init__(self, main: nn.Module, shortcut: nn.Module, drop_path: float) -> None:
        super().__init__()
        self.main = main
        self.shortcut = shortcut
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        res = self.drop_path(self.main(x))
        res = res + self.shortcut(x)

        return res


class DSConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        norm_layer: tuple[Optional[Callable[..., nn.Module]], Optional[Callable[..., nn.Module]]],
        act_layer: tuple[Optional[Callable[..., nn.Module]], Optional[Callable[..., nn.Module]]],
    ) -> None:
        super().__init__()
        self.depth_conv = Conv2dNormActivation(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            groups=in_channels,
            norm_layer=norm_layer[0],
            activation_layer=act_layer[0],
            inplace=None,
        )
        self.point_conv = Conv2dNormActivation(
            in_channels,
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            norm_layer=norm_layer[1],
            activation_layer=act_layer[1],
            inplace=None,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.depth_conv(x)
        x = self.point_conv(x)

        return x


class ConvBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        norm_layer: tuple[Optional[Callable[..., nn.Module]], Optional[Callable[..., nn.Module]]],
        act_layer: tuple[Optional[Callable[..., nn.Module]], Optional[Callable[..., nn.Module]]],
    ) -> None:
        super().__init__()
        mid_channels = in_channels

        self.conv1 = Conv2dNormActivation(
            in_channels,
            mid_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            norm_layer=norm_layer[0],
            activation_layer=act_layer[0],
            inplace=None,
        )
        self.conv2 = Conv2dNormActivation(
            mid_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=(1, 1),
            padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            norm_layer=norm_layer[1],
            activation_layer=act_layer[1],
            inplace=None,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)

        return x


class MBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        expand_ratio: float,
        norm_layer: tuple[Optional[Callable[..., nn.Module]], ...],
        act_layer: tuple[Optional[Callable[..., nn.Module]], ...],
    ) -> None:
        super().__init__()
        mid_channels = round(in_channels * expand_ratio)

        self.inverted_conv = Conv2dNormActivation(
            in_channels,
            mid_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            norm_layer=norm_layer[0],
            activation_layer=act_layer[0],
            inplace=None,
        )
        self.depth_conv = Conv2dNormActivation(
            mid_channels,
            mid_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            groups=mid_channels,
            norm_layer=norm_layer[1],
            activation_layer=act_layer[1],
            inplace=None,
        )
        self.point_conv = Conv2dNormActivation(
            mid_channels,
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            norm_layer=norm_layer[2],
            activation_layer=act_layer[2],
            inplace=None,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.inverted_conv(x)
        x = self.depth_conv(x)
        x = self.point_conv(x)

        return x


class FusedMBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        expand_ratio: float,
        norm_layer: tuple[Optional[Callable[..., nn.Module]], Optional[Callable[..., nn.Module]]],
        act_layer: tuple[Optional[Callable[..., nn.Module]], Optional[Callable[..., nn.Module]]],
    ) -> None:
        super().__init__()
        mid_channels = round(in_channels * expand_ratio)

        self.spatial_conv = Conv2dNormActivation(
            in_channels,
            mid_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            norm_layer=norm_layer[0],
            activation_layer=act_layer[0],
            inplace=None,
        )
        self.point_conv = Conv2dNormActivation(
            mid_channels,
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            norm_layer=norm_layer[1],
            activation_layer=act_layer[1],
            inplace=None,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.spatial_conv(x)
        x = self.point_conv(x)

        return x


class LiteMLA(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        heads_ratio: float,
        dim: int,
        norm_layer: tuple[Optional[Callable[..., nn.Module]], Optional[Callable[..., nn.Module]]],
        scales: list[int],
        eps: float,
    ) -> None:
        super().__init__()
        self.eps = eps
        heads = int(in_channels // dim * heads_ratio)
        total_dim = heads * dim

        self.dim = dim
        self.qkv = Conv2dNormActivation(
            in_channels,
            3 * total_dim,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            norm_layer=norm_layer[0],
            activation_layer=None,
            inplace=None,
            bias=False,
        )
        self.aggreg = nn.ModuleList(
            [
                nn.Sequential(
                    nn.Conv2d(
                        3 * total_dim,
                        3 * total_dim,
                        kernel_size=(scale, scale),
                        stride=(1, 1),
                        padding=(scale // 2, scale // 2),
                        groups=3 * total_dim,
                        bias=False,
                    ),
                    nn.Conv2d(
                        3 * total_dim,
                        3 * total_dim,
                        kernel_size=(1, 1),
                        stride=(1, 1),
                        padding=(0, 0),
                        groups=3 * heads,
                        bias=False,
                    ),
                )
                for scale in scales
            ]
        )
        self.kernel_func = nn.ReLU(inplace=False)

        self.proj = Conv2dNormActivation(
            total_dim * (1 + len(scales)),
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            norm_layer=norm_layer[1],
            activation_layer=None,
            inplace=None,
            bias=False,
        )

    def _attn(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor) -> torch.Tensor:
        dtype = v.dtype
        q = q.float()
        k = k.float()
        v = v.float()
        kv = k.transpose(-1, -2) @ v
        out = q @ kv
        out = out[..., :-1] / (out[..., -1:] + self.eps)
        return out.to(dtype)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, _, H, W) = x.size()

        # Generate multi-scale q, k, v
        qkv = self.qkv(x)
        multi_scale_qkv_list = [qkv]
        for op in self.aggreg:
            multi_scale_qkv_list.append(op(qkv))

        multi_scale_qkv = torch.concat(multi_scale_qkv_list, dim=1)
        multi_scale_qkv = multi_scale_qkv.reshape(B, -1, 3 * self.dim, H * W).transpose(-1, -2)
        (q, k, v) = multi_scale_qkv.chunk(3, dim=-1)

        # Lightweight global attention
        q = self.kernel_func(q)
        k = self.kernel_func(k)
        v = F.pad(v, (0, 1), mode="constant", value=1.0)

        with torch.autocast(device_type=v.device.type, enabled=False):
            out = self._attn(q, k, v)

        # Final projection
        out = out.transpose(-1, -2).reshape(B, -1, H, W)
        out = self.proj(out)

        return out


class EfficientVitBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        heads_ratio: float,
        head_dim: int,
        expand_ratio: float,
        norm_layer: Callable[..., nn.Module],
        act_layer: Callable[..., nn.Module],
        drop_path: float,
    ) -> None:
        super().__init__()
        self.context_module = ResidualBlock(
            LiteMLA(
                in_channels=in_channels,
                out_channels=in_channels,
                heads_ratio=heads_ratio,
                dim=head_dim,
                norm_layer=(None, norm_layer),
                scales=[5],
                eps=1e-5,
            ),
            nn.Identity(),
            drop_path=drop_path,
        )
        self.local_module = ResidualBlock(
            MBConv(
                in_channels=in_channels,
                out_channels=in_channels,
                kernel_size=(3, 3),
                stride=(1, 1),
                expand_ratio=expand_ratio,
                norm_layer=(None, None, norm_layer),
                act_layer=(act_layer, act_layer, None),
            ),
            nn.Identity(),
            drop_path=drop_path,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.context_module(x)
        x = self.local_module(x)

        return x


class EfficientVitStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        head_dim: int,
        expand_ratio: float,
        expand_divisor: int,
        norm_layer: Callable[..., nn.Module],
        act_layer: Callable[..., nn.Module],
        drop_path: list[float],
        depth: int,
        fused_block: bool,
        fewer_norms: bool,
        vit_stage: bool,
    ):
        super().__init__()
        blocks = [
            self.get_block(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=(3, 3),
                stride=(2, 2),
                expand_ratio=expand_ratio,
                norm_layer=norm_layer,
                act_layer=act_layer,
                fewer_norms=fewer_norms,
                fused_block=fused_block,
            )
        ]
        in_channels = out_channels

        if vit_stage is True:
            for i in range(depth):
                blocks.append(
                    EfficientVitBlock(
                        in_channels=in_channels,
                        heads_ratio=1.0,
                        head_dim=head_dim,
                        expand_ratio=expand_ratio // expand_divisor,
                        norm_layer=norm_layer,
                        act_layer=act_layer,
                        drop_path=drop_path[i],
                    )
                )
        else:
            for i in range(1, depth):
                blocks.append(
                    ResidualBlock(
                        self.get_block(
                            in_channels=in_channels,
                            out_channels=out_channels,
                            kernel_size=(3, 3),
                            stride=(1, 1),
                            expand_ratio=expand_ratio // expand_divisor,
                            norm_layer=norm_layer,
                            act_layer=act_layer,
                            fewer_norms=fewer_norms,
                            fused_block=fused_block,
                        ),
                        nn.Identity(),
                        drop_path=drop_path[i],
                    )
                )

        self.blocks = nn.Sequential(*blocks)

    @staticmethod
    def get_block(
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        expand_ratio: float,
        norm_layer: Callable[..., nn.Module],
        act_layer: Callable[..., nn.Module],
        fewer_norms: bool,
        fused_block: bool,
    ) -> nn.Module:
        norm_layers: tuple[Optional[Callable[..., nn.Module]], ...]

        if fused_block is False:
            if fewer_norms is True:
                norm_layers = (None, None, norm_layer)
            else:
                norm_layers = (norm_layer, norm_layer, norm_layer)

            return MBConv(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                expand_ratio=expand_ratio,
                norm_layer=norm_layers,
                act_layer=(act_layer, act_layer, None),
            )

        if fewer_norms is True:
            norm_layers = (None, norm_layer)
        else:
            norm_layers = (norm_layer, norm_layer)

        return FusedMBConv(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            expand_ratio=expand_ratio,
            norm_layer=norm_layers,
            act_layer=(act_layer, None),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.blocks(x)


class Stem(nn.Sequential):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        depth: int,
        norm_layer: Callable[..., nn.Module],
        act_layer: Callable[..., nn.Module],
        stem_block: Callable[..., nn.Module],
    ) -> None:
        super().__init__()
        self.append(
            Conv2dNormActivation(
                in_channels,
                out_channels,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                norm_layer=norm_layer,
                activation_layer=act_layer,
                inplace=None,
            ),
        )
        for _ in range(depth):
            self.append(
                ResidualBlock(
                    stem_block(
                        in_channels=out_channels,
                        out_channels=out_channels,
                        kernel_size=(3, 3),
                        stride=(1, 1),
                        norm_layer=(norm_layer, norm_layer),
                        act_layer=(act_layer, None),
                    ),
                    nn.Identity(),
                    drop_path=0.0,
                ),
            )


# pylint: disable=invalid-name
class EfficientViT_MIT(DetectorBackbone):
    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        norm_layer = nn.BatchNorm2d
        widths: list[int] = self.config["widths"]
        depths: list[int] = self.config["depths"]
        head_dim: int = self.config["head_dim"]
        conv_stages: int = self.config["conv_stages"]
        fused_blocks: int = self.config["fused_blocks"]
        head_widths: tuple[int, int] = self.config["head_widths"]
        expand_ratio: float = self.config["expand_ratio"]
        vit_expand_ratio: float = self.config["vit_expand_ratio"]
        expand_divisor: int = self.config["expand_divisor"]
        act_layer_name: str = self.config["act_layer_name"]
        stem_block_name: str = self.config["stem_block_name"]
        drop_path_rate: float = self.config["drop_path_rate"]

        act_layer = get_activation_module(act_layer_name)
        if stem_block_name == "DSConv":
            stem_block = DSConv
        elif stem_block_name == "ConvBlock":
            stem_block = ConvBlock
        else:
            raise ValueError(f"Unknown stem_block_name '{stem_block_name}'")

        self.stem = Stem(
            self.input_channels, widths[0], depths[0], norm_layer=norm_layer, act_layer=act_layer, stem_block=stem_block
        )

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths[1:])).split(depths[1:])]
        in_channels = widths[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, (w, d) in enumerate(zip(widths[1:], depths[1:])):
            stages[f"stage{i+1}"] = EfficientVitStage(
                in_channels,
                w,
                head_dim=head_dim,
                expand_ratio=vit_expand_ratio if i >= conv_stages else expand_ratio,
                expand_divisor=expand_divisor,
                norm_layer=norm_layer,
                act_layer=act_layer,
                drop_path=dpr[i],
                depth=d,
                fused_block=i < fused_blocks,
                fewer_norms=i >= 2,
                vit_stage=i >= conv_stages,
            )
            return_channels.append(w)
            in_channels = w

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                widths[-1],
                head_widths[0],
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                norm_layer=norm_layer,
                activation_layer=act_layer,
                inplace=None,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
            nn.Linear(head_widths[0], head_widths[1], bias=False),
            nn.LayerNorm(head_widths[1]),
            act_layer(),
        )
        self.return_channels = return_channels
        self.embedding_size = head_widths[1]
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "efficientvit_mit_b0",
    EfficientViT_MIT,
    config={
        "widths": [8, 16, 32, 64, 128],
        "depths": [1, 2, 2, 2, 2],
        "head_dim": 16,
        "conv_stages": 2,
        "fused_blocks": 0,
        "head_widths": (1024, 1280),
        "expand_ratio": 4.0,
        "vit_expand_ratio": 4.0,
        "expand_divisor": 1,
        "act_layer_name": "hard_swish",
        "stem_block_name": "DSConv",
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "efficientvit_mit_b1",
    EfficientViT_MIT,
    config={
        "widths": [16, 32, 64, 128, 256],
        "depths": [1, 2, 3, 3, 4],
        "head_dim": 16,
        "conv_stages": 2,
        "fused_blocks": 0,
        "head_widths": (1536, 1600),
        "expand_ratio": 4.0,
        "vit_expand_ratio": 4.0,
        "expand_divisor": 1,
        "act_layer_name": "hard_swish",
        "stem_block_name": "DSConv",
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "efficientvit_mit_b2",
    EfficientViT_MIT,
    config={
        "widths": [24, 48, 96, 192, 384],
        "depths": [1, 3, 4, 4, 6],
        "head_dim": 32,
        "conv_stages": 2,
        "fused_blocks": 0,
        "head_widths": (2304, 2560),
        "expand_ratio": 4.0,
        "vit_expand_ratio": 4.0,
        "expand_divisor": 1,
        "act_layer_name": "hard_swish",
        "stem_block_name": "DSConv",
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "efficientvit_mit_b3",
    EfficientViT_MIT,
    config={
        "widths": [32, 64, 128, 256, 512],
        "depths": [1, 4, 6, 6, 9],
        "head_dim": 32,
        "conv_stages": 2,
        "fused_blocks": 0,
        "head_widths": (2304, 2560),
        "expand_ratio": 4.0,
        "vit_expand_ratio": 4.0,
        "expand_divisor": 1,
        "act_layer_name": "hard_swish",
        "stem_block_name": "DSConv",
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "efficientvit_mit_l1",
    EfficientViT_MIT,
    config={
        "widths": [32, 64, 128, 256, 512],
        "depths": [1, 2, 2, 7, 6],
        "head_dim": 32,
        "conv_stages": 3,
        "fused_blocks": 2,
        "head_widths": (3072, 3200),
        "expand_ratio": 16.0,
        "vit_expand_ratio": 24.0,
        "expand_divisor": 4,
        "act_layer_name": "gelu_tanh",
        "stem_block_name": "ConvBlock",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "efficientvit_mit_l2",
    EfficientViT_MIT,
    config={
        "widths": [32, 64, 128, 256, 512],
        "depths": [1, 3, 3, 9, 8],
        "head_dim": 32,
        "conv_stages": 3,
        "fused_blocks": 2,
        "head_widths": (3072, 3200),
        "expand_ratio": 16.0,
        "vit_expand_ratio": 24.0,
        "expand_divisor": 4,
        "act_layer_name": "gelu_tanh",
        "stem_block_name": "ConvBlock",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "efficientvit_mit_l3",
    EfficientViT_MIT,
    config={
        "widths": [64, 128, 256, 512, 1024],
        "depths": [1, 3, 3, 9, 8],
        "head_dim": 32,
        "conv_stages": 3,
        "fused_blocks": 2,
        "head_widths": (6144, 6400),
        "expand_ratio": 16.0,
        "vit_expand_ratio": 24.0,
        "expand_divisor": 4,
        "act_layer_name": "gelu_tanh",
        "stem_block_name": "ConvBlock",
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "efficientvit_mit_b0_il-common",
    {
        "description": "EfficientViT (MIT) B0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 10.0,
                "sha256": "421197ed3bef179bb52db8675a5de5aa962ae2999e1f8d5459c87db5d956d138",
            }
        },
        "net": {"network": "efficientvit_mit_b0", "tag": "il-common"},
    },
)



"""
Faster R-CNN, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/detection/faster_rcnn.py

Paper "Faster R-CNN: Towards Real-Time Object Detection with Region Proposal Networks",
https://arxiv.org/abs/1506.01497
"""

# Reference license: BSD 3-Clause

from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import MultiScaleRoIAlign
from torchvision.ops import boxes as box_ops
from torchvision.ops.feature_pyramid_network import LastLevelMaxPool

from birder.net.base import DetectorBackbone
from birder.net.detection.base import AnchorGenerator
from birder.net.detection.base import BackboneWithFPN
from birder.net.detection.base import BoxCoder
from birder.net.detection.base import DetectionBaseNet
from birder.net.detection.base import ImageList
from birder.net.detection.base import Matcher


class BalancedPositiveNegativeSampler:
    """
    This class samples batches, ensuring that they contain a fixed proportion of positives
    """

    def __init__(self, batch_size_per_image: int, positive_fraction: float) -> None:
        self.batch_size_per_image = batch_size_per_image
        self.positive_fraction = positive_fraction

    def __call__(self, matched_idxs: list[torch.Tensor]) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        """
        Parameters
        ----------
        matched_idxs
            list of tensors containing -1, 0 or positive values.
            Each tensor corresponds to a specific image.
            -1 values are ignored, 0 are considered as negatives and > 0 as positives

        Returns
        -------
        returns two lists of binary masks for each image.
        The first list contains the positive elements that were selected,
        and the second list the negative example
        """

        pos_idx = []
        neg_idx = []
        for matched_idxs_per_image in matched_idxs:
            positive = torch.where(matched_idxs_per_image >= 1)[0]
            negative = torch.where(matched_idxs_per_image == 0)[0]

            num_pos = int(self.batch_size_per_image * self.positive_fraction)

            # Protect against not enough positive examples
            num_pos = min(positive.numel(), num_pos)
            num_neg = self.batch_size_per_image - num_pos

            # Protect against not enough negative examples
            num_neg = min(negative.numel(), num_neg)

            # Randomly select positive and negative examples
            perm1 = torch.randperm(positive.numel(), device=positive.device)[:num_pos]
            perm2 = torch.randperm(negative.numel(), device=negative.device)[:num_neg]

            pos_idx_per_image = positive[perm1]
            neg_idx_per_image = negative[perm2]

            # Create binary mask from indices
            pos_idx_per_image_mask = torch.zeros_like(matched_idxs_per_image, dtype=torch.uint8)
            neg_idx_per_image_mask = torch.zeros_like(matched_idxs_per_image, dtype=torch.uint8)

            pos_idx_per_image_mask[pos_idx_per_image] = 1
            neg_idx_per_image_mask[neg_idx_per_image] = 1

            pos_idx.append(pos_idx_per_image_mask)
            neg_idx.append(neg_idx_per_image_mask)

        return (pos_idx, neg_idx)


class RPNHead(nn.Module):
    """
    Adds a simple RPN Head with classification and regression heads
    """

    def __init__(self, in_channels: int, num_anchors: int, conv_depth: int = 1) -> None:
        super().__init__()
        conv_list = []
        for _ in range(conv_depth):
            conv_list.append(
                Conv2dNormActivation(
                    in_channels,
                    in_channels,
                    kernel_size=(3, 3),
                    stride=(1, 1),
                    padding=(1, 1),
                    norm_layer=None,
                    bias=True,
                )
            )

        self.conv = nn.Sequential(*conv_list)
        self.cls_logits = nn.Conv2d(in_channels, num_anchors, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.bbox_pred = nn.Conv2d(in_channels, num_anchors * 4, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

        # Weights initialization
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.normal_(layer.weight, std=0.01)
                if layer.bias is not None:
                    nn.init.zeros_(layer.bias)

    def forward(self, x: list[torch.Tensor]) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        logits = []
        bbox_reg = []
        for feature in x:
            t = self.conv(feature)
            logits.append(self.cls_logits(t))
            bbox_reg.append(self.bbox_pred(t))
        return logits, bbox_reg


def permute_and_flatten(layer: torch.Tensor, N: int, C: int, H: int, W: int) -> torch.Tensor:
    layer = layer.view(N, -1, C, H, W)
    layer = layer.permute(0, 3, 4, 1, 2)
    layer = layer.reshape(N, -1, C)

    return layer


def concat_box_prediction_layers(
    box_cls: list[torch.Tensor], box_regression: list[torch.Tensor]
) -> tuple[torch.Tensor, torch.Tensor]:
    box_cls_flattened = []
    box_regression_flattened = []

    # For each feature level, permute the outputs to make them be in the
    # same format as the labels. Note that the labels are computed for
    # all feature levels concatenated, so we keep the same representation
    # for the objectness and the box_regression
    for box_cls_per_level, box_regression_per_level in zip(box_cls, box_regression):
        (N, AxC, H, W) = box_cls_per_level.shape  # pylint: disable=invalid-name
        Ax4 = box_regression_per_level.shape[1]  # pylint: disable=invalid-name
        A = Ax4 // 4
        C = AxC // A
        box_cls_per_level = permute_and_flatten(box_cls_per_level, N, C, H, W)
        box_cls_flattened.append(box_cls_per_level)

        box_regression_per_level = permute_and_flatten(box_regression_per_level, N, 4, H, W)
        box_regression_flattened.append(box_regression_per_level)

    # Concatenate on the first dimension (representing the feature levels), to
    # take into account the way the labels were generated (with all feature maps
    # being concatenated as well)
    box_cls = torch.concat(box_cls_flattened, dim=1).flatten(0, -2)
    box_regression = torch.concat(box_regression_flattened, dim=1).reshape(-1, 4)

    return (box_cls, box_regression)


class RegionProposalNetwork(nn.Module):
    def __init__(
        self,
        anchor_generator: AnchorGenerator,
        head: nn.Module,
        # Faster-RCNN Training
        fg_iou_thresh: float,
        bg_iou_thresh: float,
        batch_size_per_image: int,
        positive_fraction: float,
        # Faster-RCNN Inference
        pre_nms_top_n: dict[str, int],
        post_nms_top_n: dict[str, int],
        nms_thresh: float,
        score_thresh: float = 0.0,
    ) -> None:
        super().__init__()
        self.anchor_generator = anchor_generator
        self.head = head
        self.box_coder = BoxCoder(weights=(1.0, 1.0, 1.0, 1.0))

        # Used during training
        self.box_similarity = box_ops.box_iou

        self.proposal_matcher = Matcher(
            fg_iou_thresh,
            bg_iou_thresh,
            allow_low_quality_matches=True,
        )
        self.fg_bg_sampler = BalancedPositiveNegativeSampler(batch_size_per_image, positive_fraction)

        # Used during inference
        self._pre_nms_top_n = pre_nms_top_n
        self._post_nms_top_n = post_nms_top_n
        self.nms_thresh = nms_thresh
        self.score_thresh = score_thresh
        self.min_size = 1e-3

    def pre_nms_top_n(self) -> int:
        if self.training is True:
            return self._pre_nms_top_n["training"]

        return self._pre_nms_top_n["testing"]

    def post_nms_top_n(self) -> int:
        if self.training is True:
            return self._post_nms_top_n["training"]

        return self._post_nms_top_n["testing"]

    def assign_targets_to_anchors(
        self, anchors: list[torch.Tensor], targets: list[dict[str, torch.Tensor]]
    ) -> tuple[list[torch.Tensor], list[torch.Tensor]]:

        labels = []
        matched_gt_boxes = []
        for anchors_per_image, targets_per_image in zip(anchors, targets):
            gt_boxes = targets_per_image["boxes"]

            if gt_boxes.numel() == 0:
                # Background image (negative example)
                device = anchors_per_image.device
                matched_gt_boxes_per_image = torch.zeros(anchors_per_image.shape, dtype=torch.float32, device=device)
                labels_per_image = torch.zeros((anchors_per_image.shape[0],), dtype=torch.float32, device=device)

            else:
                match_quality_matrix = self.box_similarity(gt_boxes, anchors_per_image)
                matched_idxs = self.proposal_matcher(match_quality_matrix)

                # Get the targets corresponding GT for each proposal
                # NB: need to clamp the indices because we can have a single
                # GT in the image, and matched_idxs can be -2, which goes out of bounds
                matched_gt_boxes_per_image = gt_boxes[matched_idxs.clamp(min=0)]

                labels_per_image = matched_idxs >= 0
                labels_per_image = labels_per_image.to(dtype=torch.float32)

                # Background (negative examples)
                bg_indices = matched_idxs == self.proposal_matcher.BELOW_LOW_THRESHOLD
                labels_per_image[bg_indices] = 0.0

                # Discard indices that are between thresholds
                idxs_to_discard = matched_idxs == self.proposal_matcher.BETWEEN_THRESHOLDS
                labels_per_image[idxs_to_discard] = -1.0

            labels.append(labels_per_image)
            matched_gt_boxes.append(matched_gt_boxes_per_image)

        return (labels, matched_gt_boxes)

    def _get_top_n_idx(self, objectness: torch.Tensor, num_anchors_per_level: list[int]) -> torch.Tensor:
        r = []
        offset = 0
        for ob in objectness.split(num_anchors_per_level, 1):
            num_anchors = ob.shape[1]
            pre_nms_top_n = min(self.pre_nms_top_n(), int(ob.size(1)))
            (_, top_n_idx) = ob.topk(pre_nms_top_n, dim=1)
            r.append(top_n_idx + offset)
            offset += num_anchors

        return torch.concat(r, dim=1)

    def filter_proposals(
        self,
        proposals: torch.Tensor,
        objectness: torch.Tensor,
        image_shapes: list[tuple[int, int]],
        num_anchors_per_level: list[int],
    ) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        num_images = proposals.shape[0]
        device = proposals.device

        # Do not backprop through objectness
        objectness = objectness.detach()
        objectness = objectness.reshape(num_images, -1)

        level_list: list[torch.Tensor] = [
            torch.full((n,), idx, dtype=torch.int64, device=device) for idx, n in enumerate(num_anchors_per_level)
        ]
        levels = torch.concat(level_list, dim=0)
        levels = levels.reshape(1, -1).expand_as(objectness)

        # Select top_n boxes independently per level before applying nms
        top_n_idx = self._get_top_n_idx(objectness, num_anchors_per_level)

        image_range = torch.arange(num_images, device=device)
        batch_idx = image_range[:, None]

        objectness = objectness[batch_idx, top_n_idx]
        levels = levels[batch_idx, top_n_idx]
        proposals = proposals[batch_idx, top_n_idx]

        objectness_prob = torch.sigmoid(objectness)

        final_boxes = []
        final_scores = []
        for boxes, scores, lvl, img_shape in zip(proposals, objectness_prob, levels, image_shapes):
            boxes = box_ops.clip_boxes_to_image(boxes, img_shape)

            # Remove small boxes
            keep = box_ops.remove_small_boxes(boxes, self.min_size)
            (boxes, scores, lvl) = boxes[keep], scores[keep], lvl[keep]

            # Remove low scoring boxes
            # use >= for Backwards compatibility
            keep = torch.where(scores >= self.score_thresh)[0]
            (boxes, scores, lvl) = boxes[keep], scores[keep], lvl[keep]

            # Non-maximum suppression, independently done per level
            keep = box_ops.batched_nms(boxes, scores, lvl, self.nms_thresh)

            # Keep only topk scoring predictions
            keep = keep[: self.post_nms_top_n()]
            (boxes, scores) = boxes[keep], scores[keep]

            final_boxes.append(boxes)
            final_scores.append(scores)

        return (final_boxes, final_scores)

    def compute_loss(
        self,
        objectness: torch.Tensor,
        pred_bbox_deltas: torch.Tensor,
        labels: list[torch.Tensor],
        regression_targets: list[torch.Tensor],
    ) -> tuple[torch.Tensor, torch.Tensor]:
        (sampled_pos_idxs, sampled_neg_idxs) = self.fg_bg_sampler(labels)
        sampled_pos_idxs = torch.where(torch.concat(sampled_pos_idxs, dim=0))[0]
        sampled_neg_idxs = torch.where(torch.concat(sampled_neg_idxs, dim=0))[0]

        sampled_idxs = torch.concat([sampled_pos_idxs, sampled_neg_idxs], dim=0)

        objectness = objectness.flatten()
        labels = torch.concat(labels, dim=0)

        box_loss = F.smooth_l1_loss(
            pred_bbox_deltas[sampled_pos_idxs],
            torch.concat(regression_targets, dim=0)[sampled_pos_idxs],
            beta=1 / 9,
            reduction="sum",
        ) / (sampled_idxs.numel())

        objectness_loss = F.binary_cross_entropy_with_logits(objectness[sampled_idxs], labels[sampled_idxs])

        return (objectness_loss, box_loss)

    def forward(
        self,
        images: ImageList,
        features: dict[str, torch.Tensor],
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
    ) -> tuple[list[torch.Tensor], dict[str, torch.Tensor]]:
        # RPN uses all feature maps that are available
        features_list = list(features.values())
        (objectness, pred_bbox_deltas) = self.head(features_list)
        anchors = self.anchor_generator(images, features_list)

        num_images = len(anchors)
        num_anchors_per_level_shape_tensors = [o[0].shape for o in objectness]
        num_anchors_per_level = [s[0] * s[1] * s[2] for s in num_anchors_per_level_shape_tensors]
        (objectness, pred_bbox_deltas) = concat_box_prediction_layers(objectness, pred_bbox_deltas)

        # Apply pred_bbox_deltas to anchors to obtain the decoded proposals
        # note that we detach the deltas because Faster R-CNN do not backprop through
        # the proposals
        proposals = self.box_coder.decode(pred_bbox_deltas.detach(), anchors)
        proposals = proposals.view(num_images, -1, 4)
        (boxes, _scores) = self.filter_proposals(proposals, objectness, images.image_sizes, num_anchors_per_level)

        losses: dict[str, torch.Tensor] = {}
        if self.training is True:
            if targets is None:
                raise ValueError("targets should not be None")

            (labels, matched_gt_boxes) = self.assign_targets_to_anchors(anchors, targets)
            regression_targets = self.box_coder.encode(matched_gt_boxes, anchors)
            (loss_objectness, loss_rpn_box_reg) = self.compute_loss(
                objectness, pred_bbox_deltas, labels, regression_targets
            )
            losses = {
                "loss_objectness": loss_objectness,
                "loss_rpn_box_reg": loss_rpn_box_reg,
            }

        return (boxes, losses)


class FastRCNNConvFCHead(nn.Sequential):
    def __init__(
        self,
        input_size: tuple[int, int, int],
        conv_layers: list[int],
        fc_layers: list[int],
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ):
        (in_channels, in_height, in_width) = input_size

        blocks = []
        previous_channels = in_channels
        for current_channels in conv_layers:
            blocks.append(
                Conv2dNormActivation(
                    previous_channels,
                    current_channels,
                    kernel_size=(3, 3),
                    stride=(1, 1),
                    padding=(1, 1),
                    norm_layer=norm_layer,
                    bias=False,
                )
            )
            previous_channels = current_channels

        blocks.append(nn.Flatten())
        previous_channels = previous_channels * in_height * in_width
        for current_channels in fc_layers:
            blocks.append(nn.Linear(previous_channels, current_channels))
            blocks.append(nn.ReLU(inplace=True))
            previous_channels = current_channels

        super().__init__(*blocks)

        # Weights initialization
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.kaiming_normal_(layer.weight, mode="fan_out", nonlinearity="relu")
                if layer.bias is not None:
                    nn.init.zeros_(layer.bias)


class FastRCNNPredictor(nn.Module):
    """
    Standard classification + bounding box regression layers for Fast R-CNN
    """

    def __init__(self, in_channels: int, num_classes: int) -> None:
        super().__init__()
        self.cls_score = nn.Linear(in_channels, num_classes)
        self.bbox_pred = nn.Linear(in_channels, num_classes * 4)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        if x.dim() == 4:
            torch._assert(  # pylint: disable=protected-access
                list(x.shape[2:]) == [1, 1],
                "x has the wrong shape, expecting the last two dimensions "
                f"to be [1,1] instead of {list(x.shape[2:])}",
            )

        x = x.flatten(start_dim=1)
        scores = self.cls_score(x)
        bbox_deltas = self.bbox_pred(x)

        return (scores, bbox_deltas)


def faster_rcnn_loss(
    class_logits: torch.Tensor,
    box_regression: torch.Tensor,
    label_list: list[torch.Tensor],
    regression_targets: list[torch.Tensor],
) -> tuple[torch.Tensor, torch.Tensor]:
    labels = torch.concat(label_list, dim=0)
    regression_targets = torch.concat(regression_targets, dim=0)

    classification_loss = F.cross_entropy(class_logits, labels)

    # Get indices that correspond to the regression targets for
    # the corresponding ground truth labels, to be used with
    # advanced indexing
    sampled_pos_idxs_subset = torch.where(labels > 0)[0]
    labels_pos = labels[sampled_pos_idxs_subset]
    (N, _num_classes) = class_logits.shape
    box_regression = box_regression.reshape(N, box_regression.size(-1) // 4, 4)

    box_loss = F.smooth_l1_loss(
        box_regression[sampled_pos_idxs_subset, labels_pos],
        regression_targets[sampled_pos_idxs_subset],
        beta=1 / 9,
        reduction="sum",
    )
    box_loss = box_loss / labels.numel()

    return (classification_loss, box_loss)


class RoIHeads(nn.Module):
    def __init__(
        self,
        box_roi_pool: MultiScaleRoIAlign,
        box_head: nn.Module,
        box_predictor: FastRCNNPredictor,
        # Faster R-CNN training
        fg_iou_thresh: float,
        bg_iou_thresh: float,
        batch_size_per_image: int,
        positive_fraction: float,
        bbox_reg_weights: tuple[float, float, float, float],
        # Faster R-CNN inference
        score_thresh: float,
        nms_thresh: float,
        detections_per_img: int,
        # Other
        export_mode: bool = False,
    ) -> None:
        super().__init__()

        self.box_similarity = box_ops.box_iou

        # Assign ground-truth boxes for each proposal
        self.proposal_matcher = Matcher(fg_iou_thresh, bg_iou_thresh, allow_low_quality_matches=False)

        self.fg_bg_sampler = BalancedPositiveNegativeSampler(batch_size_per_image, positive_fraction)
        self.box_coder = BoxCoder(bbox_reg_weights)

        self.box_roi_pool = box_roi_pool
        self.box_head = box_head
        self.box_predictor = box_predictor

        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.detections_per_img = detections_per_img

        if export_mode is False:
            self.forward = torch.compiler.disable(recursive=False)(self.forward)  # type: ignore[method-assign]

    def assign_targets_to_proposals(
        self, proposals: list[torch.Tensor], gt_boxes: list[torch.Tensor], gt_labels: list[torch.Tensor]
    ) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        matched_idxs = []
        labels = []
        for proposals_in_image, gt_boxes_in_image, gt_labels_in_image in zip(proposals, gt_boxes, gt_labels):
            if gt_boxes_in_image.numel() == 0:
                # Background image
                device = proposals_in_image.device
                clamped_matched_idxs_in_image = torch.zeros(
                    (proposals_in_image.shape[0],), dtype=torch.int64, device=device
                )
                labels_in_image = torch.zeros((proposals_in_image.shape[0],), dtype=torch.int64, device=device)

            else:
                # Set to self.box_similarity when https://github.com/pytorch/pytorch/issues/27495 lands
                match_quality_matrix = box_ops.box_iou(gt_boxes_in_image, proposals_in_image)
                matched_idxs_in_image = self.proposal_matcher(match_quality_matrix)

                clamped_matched_idxs_in_image = matched_idxs_in_image.clamp(min=0)

                labels_in_image = gt_labels_in_image[clamped_matched_idxs_in_image]
                labels_in_image = labels_in_image.to(dtype=torch.int64)

                # Label background (below the low threshold)
                bg_idxs = matched_idxs_in_image == self.proposal_matcher.BELOW_LOW_THRESHOLD
                labels_in_image[bg_idxs] = 0

                # Label ignore proposals (between low and high thresholds)
                ignore_idxs = matched_idxs_in_image == self.proposal_matcher.BETWEEN_THRESHOLDS
                labels_in_image[ignore_idxs] = -1  # -1 is ignored by sampler

            matched_idxs.append(clamped_matched_idxs_in_image)
            labels.append(labels_in_image)

        return (matched_idxs, labels)

    def subsample(self, labels: list[torch.Tensor]) -> list[torch.Tensor]:
        (sampled_pos_idxs, sampled_neg_idxs) = self.fg_bg_sampler(labels)
        sampled_idxs = []
        for pos_idxs_img, neg_idxs_img in zip(sampled_pos_idxs, sampled_neg_idxs):
            img_sampled_idxs = torch.where(pos_idxs_img | neg_idxs_img)[0]
            sampled_idxs.append(img_sampled_idxs)

        return sampled_idxs

    def add_gt_proposals(self, proposals: list[torch.Tensor], gt_boxes: list[torch.Tensor]) -> list[torch.Tensor]:
        proposals = [torch.concat((proposal, gt_box)) for proposal, gt_box in zip(proposals, gt_boxes)]

        return proposals

    @torch.jit.unused  # type: ignore[misc]
    @torch.compiler.disable()  # type: ignore[misc]
    def select_training_samples(
        self,
        proposals: list[torch.Tensor],
        targets: Optional[list[dict[str, torch.Tensor]]],
    ) -> tuple[list[torch.Tensor], list[torch.Tensor], list[torch.Tensor], list[torch.Tensor]]:
        if targets is None:
            raise ValueError("targets should not be None")
        if all("boxes" in t for t in targets) is False:
            raise ValueError("Every element of targets should have a boxes key")
        if all("labels" in t for t in targets) is False:
            raise ValueError("Every element of targets should have a labels key")

        dtype = proposals[0].dtype
        device = proposals[0].device

        gt_boxes = [t["boxes"].to(dtype) for t in targets]
        gt_labels = [t["labels"] for t in targets]

        # Append ground-truth bounding boxes to proposals
        proposals = self.add_gt_proposals(proposals, gt_boxes)

        # Get matching gt indices for each proposal
        (matched_idxs, labels) = self.assign_targets_to_proposals(proposals, gt_boxes, gt_labels)

        # Sample a fixed proportion of positive-negative proposals
        sampled_idxs = self.subsample(labels)
        matched_gt_boxes = []
        num_images = len(proposals)
        for img_id in range(num_images):
            img_sampled_idxs = sampled_idxs[img_id]
            proposals[img_id] = proposals[img_id][img_sampled_idxs]
            labels[img_id] = labels[img_id][img_sampled_idxs]
            matched_idxs[img_id] = matched_idxs[img_id][img_sampled_idxs]

            gt_boxes_in_image = gt_boxes[img_id]
            if gt_boxes_in_image.numel() == 0:
                gt_boxes_in_image = torch.zeros((1, 4), dtype=dtype, device=device)

            matched_gt_boxes.append(gt_boxes_in_image[matched_idxs[img_id]])

        regression_targets = self.box_coder.encode(matched_gt_boxes, proposals)

        return (proposals, matched_idxs, labels, regression_targets)

    def postprocess_detections(
        self,
        class_logits: torch.Tensor,
        box_regression: torch.Tensor,
        proposals: list[torch.Tensor],
        image_shapes: list[tuple[int, int]],
    ) -> tuple[list[torch.Tensor], list[torch.Tensor], list[torch.Tensor]]:
        device = class_logits.device
        num_classes = class_logits.shape[-1]

        boxes_per_image = [boxes_in_image.shape[0] for boxes_in_image in proposals]
        pred_boxes = self.box_coder.decode(box_regression, proposals)

        pred_scores = F.softmax(class_logits, -1)

        pred_boxes_list = pred_boxes.split(boxes_per_image, 0)
        pred_scores_list = pred_scores.split(boxes_per_image, 0)

        all_boxes = []
        all_scores = []
        all_labels = []
        for boxes, scores, image_shape in zip(pred_boxes_list, pred_scores_list, image_shapes):
            boxes = box_ops.clip_boxes_to_image(boxes, image_shape)

            # Create labels for each prediction
            labels = torch.arange(num_classes, device=device)
            labels = labels.view(1, -1).expand_as(scores)

            # Remove predictions with the background label
            boxes = boxes[:, 1:]
            scores = scores[:, 1:]
            labels = labels[:, 1:]

            # Batch everything, by making every class prediction be a separate instance
            boxes = boxes.reshape(-1, 4)
            scores = scores.reshape(-1)
            labels = labels.reshape(-1)

            # Remove low scoring boxes
            idxs = torch.where(scores > self.score_thresh)[0]
            boxes = boxes[idxs]
            scores = scores[idxs]
            labels = labels[idxs]

            # Remove empty boxes
            keep = box_ops.remove_small_boxes(boxes, min_size=1e-2)
            boxes = boxes[keep]
            scores = scores[keep]
            labels = labels[keep]

            # Non-maximum suppression, independently done per class
            keep = box_ops.batched_nms(boxes, scores, labels, self.nms_thresh)

            # Keep only topk scoring predictions
            keep = keep[: self.detections_per_img]
            boxes = boxes[keep]
            scores = scores[keep]
            labels = labels[keep]

            all_boxes.append(boxes)
            all_scores.append(scores)
            all_labels.append(labels)

        return (all_boxes, all_scores, all_labels)

    def forward(  # pylint: disable=method-hidden
        self,
        features: dict[str, torch.Tensor],
        proposals: list[torch.Tensor],
        image_shapes: list[tuple[int, int]],
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        if targets is not None:
            for t in targets:
                floating_point_types = (torch.float, torch.double, torch.half)
                if t["boxes"].dtype not in floating_point_types:
                    raise TypeError(f"target boxes must of float type, instead got {t['boxes'].dtype}")
                if t["labels"].dtype != torch.int64:
                    raise TypeError(f"target labels must of int64 type, instead got {t['labels'].dtype}")

        if self.training is True:
            (proposals, _matched_idxs, labels, regression_targets) = self.select_training_samples(proposals, targets)
        else:
            labels = None
            regression_targets = None
            _matched_idxs = None  # noqa: F841

        box_features = self.box_roi_pool(features, proposals, image_shapes)
        box_features = self.box_head(box_features)
        (class_logits, box_regression) = self.box_predictor(box_features)

        losses = {}
        result: list[dict[str, torch.Tensor]] = []
        if self.training is True:
            if labels is None:
                raise ValueError("labels cannot be None")
            if regression_targets is None:
                raise ValueError("regression_targets cannot be None")

            (loss_classifier, loss_box_reg) = faster_rcnn_loss(class_logits, box_regression, labels, regression_targets)
            losses = {"loss_classifier": loss_classifier, "loss_box_reg": loss_box_reg}

        else:
            (boxes, scores, labels) = self.postprocess_detections(class_logits, box_regression, proposals, image_shapes)
            num_images = len(boxes)
            for i in range(num_images):
                result.append(
                    {
                        "boxes": boxes[i],
                        "labels": labels[i],
                        "scores": scores[i],
                    }
                )

        return (result, losses)


# pylint: disable=invalid-name
class Faster_RCNN(DetectionBaseNet):
    default_size = (640, 640)
    auto_register = True

    # pylint: disable=too-many-locals
    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is None, "config not supported"

        fpn_width = 256
        rpn_pre_nms_top_n_train = 2000
        rpn_pre_nms_top_n_test = 1000
        rpn_post_nms_top_n_train = 2000
        rpn_post_nms_top_n_test = 1000
        rpn_fg_iou_thresh = 0.7
        rpn_bg_iou_thresh = 0.3
        rpn_batch_size_per_image = 256
        rpn_positive_fraction = 0.5
        rpn_nms_thresh = 0.7
        rpn_score_thresh = 0.0
        box_fg_iou_thresh = 0.5
        box_bg_iou_thresh = 0.5
        box_batch_size_per_image = 512
        box_positive_fraction = 0.25
        bbox_reg_weights = (10.0, 10.0, 5.0, 5.0)
        box_score_thresh = 0.05
        box_nms_thresh = 0.5
        box_detections_per_img = 100
        canonical_scale = 224

        self.backbone_with_fpn = BackboneWithFPN(self.backbone, fpn_width, extra_blocks=LastLevelMaxPool())

        anchor_sizes = [[32], [64], [128], [256], [512]]
        anchor_sizes = anchor_sizes[-len(self.backbone.return_stages) - 1 :]
        aspect_ratios = [[0.5, 1.0, 2.0]] * len(anchor_sizes)
        rpn_anchor_generator = AnchorGenerator(anchor_sizes, aspect_ratios)
        rpn_head = RPNHead(
            self.backbone_with_fpn.out_channels, rpn_anchor_generator.num_anchors_per_location()[0], conv_depth=2
        )

        rpn_pre_nms_top_n = {"training": rpn_pre_nms_top_n_train, "testing": rpn_pre_nms_top_n_test}
        rpn_post_nms_top_n = {"training": rpn_post_nms_top_n_train, "testing": rpn_post_nms_top_n_test}

        self.rpn = RegionProposalNetwork(
            rpn_anchor_generator,
            rpn_head,
            rpn_fg_iou_thresh,
            rpn_bg_iou_thresh,
            rpn_batch_size_per_image,
            rpn_positive_fraction,
            rpn_pre_nms_top_n,
            rpn_post_nms_top_n,
            rpn_nms_thresh,
            score_thresh=rpn_score_thresh,
        )
        box_roi_pool = MultiScaleRoIAlign(
            featmap_names=self.backbone.return_stages,
            output_size=canonical_scale // 32,
            sampling_ratio=2,
            canonical_scale=canonical_scale,
        )
        box_head = FastRCNNConvFCHead(
            (self.backbone_with_fpn.out_channels, canonical_scale // 32, canonical_scale // 32),
            [256, 256, 256, 256],
            [1024],
            norm_layer=nn.BatchNorm2d,
        )

        self.representation_size = 1024
        box_predictor = FastRCNNPredictor(self.representation_size, self.num_classes)

        self.roi_heads = RoIHeads(
            box_roi_pool,
            box_head,
            box_predictor,
            box_fg_iou_thresh,
            box_bg_iou_thresh,
            box_batch_size_per_image,
            box_positive_fraction,
            bbox_reg_weights,
            box_score_thresh,
            box_nms_thresh,
            box_detections_per_img,
            export_mode=self.export_mode,
        )

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes + 1
        box_predictor = FastRCNNPredictor(self.representation_size, self.num_classes)
        self.roi_heads.box_predictor = box_predictor

    def freeze(self, freeze_classifier: bool = True) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.roi_heads.box_predictor.parameters():
                param.requires_grad = True

    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[list[int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        self._input_check(targets)
        images = self._to_img_list(x, image_sizes)

        features = self.backbone_with_fpn(x)
        (proposals, proposal_losses) = self.rpn(images, features, targets)
        (detections, detector_losses) = self.roi_heads(features, proposals, images.image_sizes, targets)

        losses = {}
        losses.update(detector_losses)
        losses.update(proposal_losses)

        return (detections, losses)



"""
LARS, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/optim/lars.py

Paper "Large Batch Training of Convolutional Networks", https://arxiv.org/abs/1708.03888
"""

# Reference license: Apache-2.0

from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
from torch.optim.optimizer import Optimizer
from torch.optim.optimizer import ParamsT


class Lars(Optimizer):
    """
    Layer-wise Adaptive Rate Scaling (LARS) for PyTorch

    Taken as-is from timm - https://github.com/huggingface/pytorch-image-models/blob/main/timm/optim/lars.py
    Type annotations added.
    """

    def __init__(
        self,
        params: ParamsT,
        lr: float,
        momentum: float = 0.0,
        dampening: float = 0.0,
        weight_decay: float = 0.0,
        nesterov: bool = False,
        trust_coeff: float = 0.001,
        eps: float = 1e-8,
        trust_clip: bool = False,
        always_adapt: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        params
            Iterable of parameters to optimize or dicts defining parameter groups.
        lr
            Learning rate.
        momentum
            Momentum factor. Enables momentum-based acceleration.
        dampening
            Dampening for momentum. Reduces the contribution of previous gradients.
        weight_decay
            Weight decay (L2 penalty).
        nesterov
            Enables Nesterov momentum.
        trust_coeff
            Trust coefficient for computing adaptive lr / trust_ratio.
            Controls how aggressively to adapt the learning rate.
        eps
            Small constant added to denominator for numerical stability.
        trust_clip
            When True, enables clipping of the LARS trust ratio to prevent extreme values.
        always_adapt
            When True, always applies LARS learning rate adaptation.
            When False, only applies adaptation when group weight_decay != 0.
        """

        if lr < 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if momentum < 0.0:
            raise ValueError(f"Invalid momentum value: {momentum}")
        if weight_decay < 0.0:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")

        defaults = {
            "lr": lr,
            "momentum": momentum,
            "dampening": dampening,
            "weight_decay": weight_decay,
            "nesterov": nesterov,
            "trust_coeff": trust_coeff,
            "eps": eps,
            "trust_clip": trust_clip,
            "always_adapt": always_adapt,
        }
        super().__init__(params, defaults)

    def __setstate__(self, state: dict[str, Any]) -> None:
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("nesterov", False)

    @torch.no_grad()  # type: ignore[misc]
    def step(self, closure: Optional[Callable[[], float]] = None) -> Optional[float]:
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            weight_decay = group["weight_decay"]
            momentum = group["momentum"]
            dampening = group["dampening"]
            nesterov = group["nesterov"]
            trust_coeff = group["trust_coeff"]
            eps = group["eps"]

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad

                # Apply LARS LR adaptation, LARC clipping, weight decay
                # ref: https://github.com/NVIDIA/apex/blob/master/apex/parallel/LARC.py
                if weight_decay != 0 or group["always_adapt"] is True:
                    w_norm = p.norm(2.0)
                    g_norm = grad.norm(2.0)
                    trust_ratio = trust_coeff * w_norm / (g_norm + w_norm * weight_decay + eps)
                    # FIX: nested where required since logical and/or not working in PT XLA
                    # Set the ratio to 1.0 (no change) if either weight norm or grad norm is zero
                    trust_ratio = torch.where(
                        w_norm > 0,
                        torch.where(g_norm > 0, trust_ratio, 1.0),
                        1.0,
                    )
                    if group["trust_clip"] is True:
                        trust_ratio = torch.clamp(trust_ratio / group["lr"], max=1.0)

                    grad.add_(p, alpha=weight_decay)
                    grad.mul_(trust_ratio)

                # Apply SGD update https://github.com/pytorch/pytorch/blob/1.7/torch/optim/sgd.py#L100
                if momentum != 0:
                    param_state = self.state[p]
                    if "momentum_buffer" not in param_state:
                        buf = param_state["momentum_buffer"] = torch.clone(grad).detach()
                    else:
                        buf = param_state["momentum_buffer"]
                        buf.mul_(momentum).add_(grad, alpha=1.0 - dampening)
                    if nesterov is True:
                        grad = grad.add(buf, alpha=momentum)
                    else:
                        grad = buf

                p.add_(grad, alpha=-group["lr"])

        return loss
